#define SCC 1
#ifdef SCC

#define BANNER "Z8530 SCC driver v1.4a.dl1bke (beta) by pe1nnz and dl1bke\n"

/************************************************************************/
/* SCC.C								*/ 
/************************************************************************/

/*
   Well... We will drop the usual history list. Among the people working
   on this driver were:
   
   PE1CHL Rob   - the original author - NET version
   KA9Q   Phil  - NOS version
   KY3B   Ken   - various changes
   DG0FT  Rene  - BayCom USCC support (for NOS)
   
   Linux version:

   PA3AOU Harry - testing, information supply and supporting
   PE1NNZ Guido - ported to Linux, ESCC support 
   DL1BKE Joerg - KISS-parameter setting and BayCom USCC support
*/

/* exports: scc_init(), scc_open(), scc_write() */
/* changes: in tty.h, tty_io.c and Makefile */
/* to do  : change scc.h to your needs */

/* BTW: why are *_write() global? */


#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/interrupt.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/scc.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <linux/kernel.h>


long scc_init(long kmem_start);
int scc_open(struct tty_struct *tty, struct file * filp);
void scc_write(struct tty_struct *tty);

static int z8530_init(int nchips,ioaddr iobase,int space,int aoff,
                    int boff,int doff,ioaddr intack,int ivec,
                    long clk,int pclk,int hwtype,int hwparam);

static void scc_close(struct tty_struct *tty, struct file * filp);
static void scc_change_speed(unsigned int line);

#ifdef oldlinux
static int scc_ioctl(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned int arg);
#else
static int scc_ioctl(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg);
#endif

static void scc_set_termios(struct tty_struct *tty, struct termios * oldtermios);
static void scc_rxraw(struct sccchan *scc);

static void scc_sdlc(struct sccchan *scc);
static void scc_tossb(struct sccchan *scc);
static void scc_txon(struct sccchan *scc);
static void scc_txoff(struct sccchan *scc);
static unsigned int scc_speed (struct sccchan *scc, unsigned int clkmode,long speed);
static void scc_sdlctx(register struct sccchan *scc);
static void scc_sdlcex(register struct sccchan *scc);
static void scc_sdlcrx(register struct sccchan *scc);
static void scc_sdlcsp(register struct sccchan *scc);
static void sccvec(int irq);
static void sccnovec(int irq);
static void scc_timer(void);

/* from serial.c */

static int baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 0 };
	
struct sccinfo Sccinfo = {0};               /* global info about SCCs */
struct sccchan *Sccchan[2 * MAXSCC] = {0};  /* information per channel */
ioaddr Sccvecloc = {0};                     /* location to access for SCC vector */
unsigned char Sccmaxvec = {0};              /* maximum legal vector from SCC */
ioaddr Sccpolltab[MAXSCC+1][2] = {0,0};     /* polling table when no vectoring */

#if defined(INLINE)
static unsigned scc_delay(unsigned v);

static unsigned scc_delay(unsigned v)    /* delay for about 5 PCLK cycles */
/* unsigned v;                              pass-through used for input */
{
   register int i,j;			/* it takes time to save them */
   return v;                    /* return the passed parameter */
}
#endif

unsigned char Random = 0;		/* random number for p-persist */


int dirps(void)
{
	cli();
	return 0;
}

void restore(int i_state)
{
	sti();
}


/* ----> some code stolen from mbuf.c <---- */


/* Allocate mbuf with associated buffer of 'size' bytes */
struct mbuf *
alloc_mbuf(register int size)
{
	register struct mbuf *bp;

	bp = (struct mbuf *) kmalloc((unsigned)(size + sizeof(struct mbuf)), GFP_ATOMIC);

	if(bp == NULLBUF)
		return NULLBUF;
	/* Clear just the header portion */
	memset((char *)bp,0,sizeof(struct mbuf));
	if((bp->size = size) != 0)
		bp->data = (char *)(bp + 1);
	bp->refcnt++;

	return bp;
}

/* Decrement the reference pointer in an mbuf. If it goes to zero,
 * free all resources associated with mbuf.
 * Return pointer to next mbuf in packet chain
 */
struct mbuf *
free_mbuf(register struct mbuf *bp)
{
	struct mbuf *bpnext;

	if(bp == NULLBUF)
		return NULLBUF;

	bpnext = bp->next;
	if(bp->dup != NULLBUF){
		free_mbuf(bp->dup);     /* Follow indirection */
		bp->dup = NULLBUF;
	}
	/* Decrement reference count. If it has gone to zero, free it. */
	if(--bp->refcnt <= 0)
		kfree((char *)bp);
		
	return bpnext;
}

/* Free packet (a chain of mbufs). Return pointer to next packet on queue,
 * if any
 */
struct mbuf *
free_p(register struct mbuf *bp)
{
	register struct mbuf *abp;

	if(bp == NULLBUF)
		return NULLBUF;
	abp = bp->anext;
	while(bp != NULLBUF)
		bp = free_mbuf(bp);
	return abp;
}               

/* Append packet to end of packet queue */
void
enqueue(struct mbuf **q,struct mbuf *bp)
{
	register struct mbuf *p;
	char i_state;

	if(q == NULLBUFP || bp == NULLBUF)
		return;
	i_state = dirps();
	if(*q == NULLBUF){
		/* List is empty, stick at front */
		*q = bp;
	} else {
		for(p = *q ; p->anext != NULLBUF ; p = p->anext)
			;
		p->anext = bp;
	}
	restore(i_state);
}

/* Append mbuf to end of mbuf chain */
void
append(struct mbuf **bph,struct mbuf *bp)
{
	register struct mbuf *p;
	char i_state;

	if(bph == NULLBUFP || bp == NULLBUF)
		return;
	
	i_state=dirps();
	if(*bph == NULLBUF){
		/* First one on chain */
		*bph = bp;
	} else {
		for(p = *bph ; p->next != NULLBUF ; p = p->next)
			;
		p->next = bp;
	}
	restore(i_state);
}


/* ---->  SCC driver initialisation. Called at boot time <---- */

long scc_init(long kmem_start)
{
	printk(BANNER);
	z8530_init(NCHIPS, IOBASE, SPACE, AOFF, BOFF, DOFF, INTACK, IVEC, CLK, PCLK, HWTYPE, HWPARAM); 
	return kmem_start;
}

/* here for historic reasons */

static int
z8530_init(int nchips,ioaddr iobase,int space,int aoff,int boff,int doff,ioaddr intack,int ivec,long clk,int pclk,int hwtype,int hwparam)
/* int nchips;			number of chips */
/* ioaddr iobase;      		base of first chip */
/* int space,aoff,boff,doff; */
/* ioaddr intack;      		INTACK ioport or 0 for no INTACK */
/* int ivec;			interrupt vector number */
/* long clk;		        clock frequency */
/* int pclk;			PCLK or RTxC for clock */
/* int hwtype;			selection of special hardware types */
/* int hwparam;			extra parameter for special hardware */
{
	struct sigaction	sa;
	int chip,chan;
	ioaddr chipbase;
	register ioaddr ctrl;
    	int i_state,d;
    	int dum = 1;

	Sccinfo.nchips = nchips;
	Sccinfo.maxchan = (2 * nchips) - 1;
	Sccinfo.iobase = iobase;
	Sccinfo.space = space;
	Sccinfo.off[0] = aoff;
	Sccinfo.off[1] = boff;
	Sccinfo.doff = doff;
	Sccinfo.ivec = ivec;
	Sccinfo.clk = clk;
	Sccinfo.pclk = pclk;
	Sccinfo.hwtype = hwtype;
	Sccinfo.hwparam = hwparam;

#define z 0

	/* reset and pre-init all chips in the system */
	for(chip = 0; chip < nchips; chip++){
		chipbase = iobase + chip * space;
		ctrl = chipbase + Sccinfo.off[0];
		i_state = dirps();		/* because of 2-step accesses */
		VOID(RDREG(ctrl));		/* make sure pointer is written */
		WRSCC(ctrl,R9,FHWRES);		/* force hardware reset */
		for (d = 0; d < 1000; d++)	/* wait a while to be sure */
			dum *= 10;
		for(chan = 0; chan < 2; chan++){
			Sccchan[2 * chip + chan] = NULLCHAN;
			ctrl = chipbase + Sccinfo.off[chan];

			/* initialize a single channel to no-op */
			VOID(RDREG(ctrl));		/* make sure pointer is written */
			WRSCC(ctrl,R4,z);		/* no mode selected yet */
			WRSCC(ctrl,R1,z);		/* no W/REQ operation */
			WRSCC(ctrl,R2,16 * chip);	/* chip# in upper 4 bits of vector */
			WRSCC(ctrl,R3,z);		/* disable rx */
			WRSCC(ctrl,R5,z);		/* disable tx */
			WRSCC(ctrl,R9,VIS);		/* vector includes status, MIE off */
			Sccpolltab[chip][chan] = ctrl;	/* store ctrl addr for polling */
		}

		/* Other SCC cards */

		if(hwtype & HWEAGLE)			/* this is an EAGLE card */
			WRREG(chipbase + 4,0x08);	/* enable interrupt on the board */

		if(hwtype & HWPC100)			/* this is a PC100 card */
			WRREG(chipbase,hwparam);	/* set the MODEM mode (22H normally) */

		if(hwtype & HWPRIMUS)			/* this is a PRIMUS-PC */
			WRREG(chipbase + 4,hwparam); 	/* set the MODEM mode (02H normally) */

		if (hwtype & HWDRSI) {			/* this is a DRSI PC*Packet card */
			ioaddr z8536 = chipbase + 7; 	/* point to 8536 master ctrl reg */

			/* Initialize 8536 to perform its divide-by-32 function */
			/* This part copied from N6TTO DRSI-driver */

			/* Start by forcing chip into known state */

			VOID(RDREG(z8536));		/* make sure pointer is written */
			WRSCC(z8536,CIO_MICR,0x01); 	/* force hardware reset */

			for (d = 0; d < 1000; d++)	/* wait a while to be sure */
				dum *= 10;

			WRSCC(z8536,CIO_MICR,0x00); 	/* Clear reset and start */

			/* Wait for chip to come ready */

			while (RDSCC(z8536,CIO_MICR) != 0x02)
				dum *= 10;

			WRSCC(z8536,CIO_MICR,0x26); /* NV|CT_VIS|RJA */
			WRSCC(z8536,CIO_MCCR,0xf4); /* PBE|CT1E|CT2E|CT3E|PAE */

			WRSCC(z8536,CIO_CTMS1,0xe2);/* Continuous, EOE, ECE, Pulse output */
			WRSCC(z8536,CIO_CTMS2,0xe2);/* Continuous, EOE, ECE, Pulse output */
			WRSCC(z8536,CIO_CT1MSB,0x00); /* Load time constant CTC #1 */
			WRSCC(z8536,CIO_CT1LSB,0x10);
			WRSCC(z8536,CIO_CT2MSB,0x00); /* Load time constant CTC #2 */
			WRSCC(z8536,CIO_CT2LSB,0x10);

			WRSCC(z8536,CIO_IVR,0x06);

			/* Set port direction bits in port A and B		     */
			/* Data is input on bits d1 and d5, output on d0 and d4. */
			/* The direction is set by 1 for input and 0 for output  */

			WRSCC(z8536,CIO_PDCA,0x22);
			WRSCC(z8536,CIO_PDCB,0x22);

			WRSCC(z8536,CIO_CSR1,CIO_GCB|CIO_TCB); /* Start CTC #1 running */
			WRSCC(z8536,CIO_CSR2,CIO_GCB|CIO_TCB); /* Start CTC #2 running */
		}
        	restore(i_state);
        }
	Sccpolltab[chip][0] = 0;        /* terminate the polling table */
	Sccvecloc = intack;             /* location of INTACK/vector read */
	Sccmaxvec = 16 * nchips;        /* upper limit on valid vector */
	/* save original interrupt vector */

	/* set interrupt vector to INTACK-generating routine  */

	if(ivec == 2)
		ivec = 9;
	if(intack)
		sa.sa_handler = sccvec;
	else
		sa.sa_handler = sccnovec;
		
	sa.sa_flags = (SA_INTERRUPT);
	sa.sa_mask = 0;
	sa.sa_restorer = NULL;
	irqaction(ivec,&sa);
	
	/* enable the interrupt  */
	/* maskon(ivec); */

	/* set timer interrupt to scctimer() */
        /* setvect(0x1c,scctimer); */
/*        
	timer_table[SCC_TIMER].fn = scc_timer;
	timer_table[SCC_TIMER].expires = 0;
	timer_active |= 1 << SCC_TIMER;
*/
	printk("Init SCC driver: channels=%u, base=0x%x, irq=%u\n",Sccinfo.nchips*2,Sccinfo.iobase,Sccinfo.ivec);

	return 0;
}


/* ----> this one is called whenever you open the device <---- */

int scc_open(struct tty_struct *tty, struct file * filp)
{
	struct tty_queue * queue;
	struct sccchan *scc;
	unsigned int chan;

	chan = DEV_TO_SL(tty->line) - SC_TTY_MASK;
	if(chan > Sccinfo.maxchan)
		return -1;
	if(Sccchan[chan] != NULLCHAN){
		scc = Sccchan[chan];
		scc->tty_opened++;
		return 0;
	}

	tty->write = scc_write;
	tty->close = scc_close;
	tty->ioctl = scc_ioctl;
	
	
/* !!!! higher patchlevels !!!! */	
#ifndef oldlinux
	tty->set_termios = scc_set_termios;
#endif	
/* --- */

	tty->throttle = 0; 
	tty->termios->c_cflag &= ~CBAUD;

	scc = (struct sccchan *) kmalloc(sizeof(struct sccchan), GFP_KERNEL);
	memset((char *)scc,0,sizeof(struct sccchan));

 	scc->tty = tty;

#ifdef oldlinux 	
/* !!!! for patchlevel <= 5 !!!! */
 	scc->tty_opened=1;
/* --- */
#endif
 
	scc->ctrl = Sccinfo.iobase + (chan / 2) * Sccinfo.space + Sccinfo.off[chan % 2];
	scc->data = scc->ctrl + Sccinfo.doff;

/* !!!! BayCom USCC has no fullduplex divider !!!! */

	scc->fulldup = SCC_Fulldup[chan];
	scc->extclock = SCC_Extclk[chan];
	
/* --- */	

	/* Enhanced SCC support */
	scc->enhanced = SCC_Enhanced[chan];

	scc_sdlc(scc);              /* init SCC in SDLC mode */

	scc->speed = Sccinfo.clk / (64L * (scc_speed(scc,32,1200) + 2));/* init SCC speed */

	scc->a.bufsiz = BUFSIZE;    /* packet buffer size */

	/* default KISS Params */
	scc->a.txdelay = 36*TPS/100;    /* 360 ms */
	scc->a.persist = 64;            /* 25% persistence */			/* was 25 */
	scc->a.slottime = 16*TPS/100;   /* 160 ms */
#if TPS > 67
	scc->a.tailtime = 3*TPS/100;    /* 30 ms */
#else
	scc->a.tailtime = 2;            /* minimal reasonable value */
#endif
	scc->a.fulldup = 0;             /* CSMA */
	scc->a.waittime = 50*TPS/100;   /* 500 ms */
	scc->a.maxkeyup = 7;            /* 7 s */
	scc->a.mintime = 3;             /* 3 s */
	scc->a.idletime = 120;          /* 120 s */
	
	scc->a.maxdefer = TPS * scc->a.idletime / scc->a.slottime;
	if (scc->a.maxdefer == 0) scc->a.maxdefer = 400;	

	Sccchan[chan] = scc;		/* put addr in table for interrupts */

	queue = &scc->tty->read_q;
	queue->buf[queue->head++] = FEND;
	queue->head &= TTY_BUF_SIZE-1;
	queue->buf[queue->head++] = 0;
	queue->head &= TTY_BUF_SIZE-1;
 	TTY_READ_FLUSH(tty);

	timer_table[SCC_TIMER].fn = scc_timer;
	timer_table[SCC_TIMER].expires = 0;
	timer_active |= 1 << SCC_TIMER;
	
	return 0;
}


/* ----> and this whenever you close the device <---- */

static void
scc_close(struct tty_struct *tty, struct file * filp)
{
	struct sccchan *scc;
	unsigned int chan;
	int i_state;

	chan = DEV_TO_SL(tty->line) - SC_TTY_MASK;
	if((scc = Sccchan[chan]) == NULLCHAN)
		return;
		
/* !!!! for patchlevel <= 5 !!!! */
#ifdef oldlinux		
	if(scc->tty_opened > 1){
#else
	if(scc->tty_opened){
#endif	
		scc->tty_opened--;
		return;
	}
/* --- */
	
	i_state = dirps();
	VOID(RDREG(scc->ctrl));		/* Make sure pointer is written */
	
        wr(scc,R1, 0);			/* disable interrupts */
        wr(scc,R3, 0);			/* disable rx */
	
/* Hmm... do this and you'll have to initialize the whole chip... */                
#ifdef 0
	wr(scc,R9,(chan % 2) ? CHRB : CHRA);	/* reset the channel */
#endif

	Sccchan[chan] = NULLCHAN;
	restore(i_state);
	tty->stopped = 0;		
	kfree(scc);
}

/* ----> a subroutine for historic reasons <---- */

/* initialize an SCC channel in SDLC mode */
static void
scc_sdlc(register struct sccchan *scc)
{
	int i_state;

    	/* set interrupt handlers */
    	scc->int_transmit = scc_sdlctx;
	scc->int_extstat = scc_sdlcex;
	scc->int_receive = scc_sdlcrx;
	scc->int_special = scc_sdlcsp;

	i_state = dirps();

	wr(scc,R4,X1CLK|SDLC);		/* *1 clock, SDLC mode */
	wr(scc,R1,z);			/* no W/REQ operation */
	wr(scc,R3,Rx8|RxCRC_ENAB);	/* RX 8 bits/char, CRC, disabled */
	wr(scc,R5,Tx8|DTR|TxCRC_ENAB);	/* TX 8 bits/char, disabled, DTR */
	wr(scc,R6,z);			/* SDLC address zero (not used) */
	wr(scc,R7,FLAG);		/* SDLC flag value */
	wr(scc,R9,VIS);			/* vector includes status */
/*	wr(scc,R10,CRCPS|NRZI|ABUNDER);	*/ /* CRC preset 1, select NRZI, ABORT on underrun */
	
	if (scc->extclock){			/* when using external clocks */
	
/* !!!! NRZ instead of NRZI for BayCom !!!! */

		if (Sccinfo.hwtype & HWBAYCOM){	/* DF9IC modem? */
			wr(scc,R10,CRCPS|NRZ|ABUNDER);	/* CRC preset 1, select NRZ, ABORT on underrun */
			wr(scc,R11,RCTRxCP|TCRTxCP);	/* RXclk, TRxC, TXclk RTxC */
		} else {	
			wr(scc,R10,CRCPS|NRZI|ABUNDER); /* CRC preset 1, select NRZI, ABORT on underrun */
			wr(scc,R11,RCRTxCP|TCTRxCP);	/* RXclk RTxC, TXclk TRxC. */   
		}
/* --- */
		
			wr(scc,R14,z);			/* No BRG options */
		WRSCC(scc->ctrl,R14,DISDPLL|scc->wreg[R14]); /* No DPLL operation */
	} else {
		wr(scc,R10,CRCPS|NRZI|ABUNDER);
		if(scc->fulldup){		/* when external clock divider */
			if(Sccinfo.pclk){	/* when using PCLK as clock source */
			
/* !!!! hmm, don't know why... !!!! */

				if (Sccinfo.hwtype & HWBAYCOM){
					/* RXclk DPLL, TXclk RTxC, out=DPLL (Echo duplex) */
					wr(scc,R11,RCDPLL|TCRTxCP|TRxCOI|TRxCDP);
				} else {
					/* RXclk DPLL, TXclk RTxC, out=BRG.	 external /32 TRxC->RTxC */
					wr(scc,R11,RCDPLL|TCRTxCP|TRxCOI|TRxCBR);
				}
/* --- */
				
				
			} else {
				/* RXclk DPLL, TXclk TRxC.	external TX clock to TRxC */
				wr(scc,R11,RCDPLL|TCTRxCP);
			}
		} else {			/* only half-duplex operation */
			/* RXclk DPLL, TXclk BRG. BRG reprogrammed at every TX/RX switch */
#ifdef	notdef	/* KA9Q - for PSK modem */
			wr(scc,R11,RCDPLL|TCBR);
#else
			/* DPLL -> Rx clk, DPLL -> Tx CLK, TxCLK -> TRxC pin */
			wr(scc,R11,RCDPLL|TCDPLL|TRxCOI|TRxCDP);
#endif
		}
		
		wr(scc,R14,Sccinfo.pclk? BRSRC:z);	/* BRG source = PCLK/RTxC */
		WRSCC(scc->ctrl,R14,SSBR|scc->wreg[R14]); /* DPLL source = BRG */
		WRSCC(scc->ctrl,R14,SNRZI|scc->wreg[R14]); /* DPLL NRZI mode */
	}
	
	/* enable CTS (not for Baycom), ABORT & DCD interrupts */
	wr(scc,R15,((Sccinfo.hwtype & HWBAYCOM) ? 0 : CTSIE)|BRKIE|DCDIE);

	if(scc->enhanced){
		or(scc,R15,SHDLCE|FIFOE);	/* enable FIFO, SDLC/HDLC Enhancements (From now R7 is R7') */
		wr(scc,R7,AUTOEOM);
	}

	if(RDREG(scc->ctrl) & DCD){	/* DCD is now ON */
		if (!scc->extclock)
			WRSCC(scc->ctrl,R14,SEARCH|scc->wreg[R14]); /* DPLL: enter search mode */
		or(scc,R3,ENT_HM|RxENABLE);	/* enable the receiver, hunt mode */
	}
	WRREG(scc->ctrl,RES_EXT_INT);	/* reset ext/status interrupts */
	WRREG(scc->ctrl,RES_EXT_INT);	/* must be done twice */
	scc->status = RDREG(scc->ctrl);	/* read initial status */

	or(scc,R1,INT_ALL_Rx|TxINT_ENAB|EXT_INT_ENAB); /* enable interrupts */
	or(scc,R9,MIE);			/* master interrupt enable */

	restore(i_state);
}

/***************************************************************************/


/* ----> set SCC channel speed <---- */

/* clkmode specifies the division rate (1,16,32) inside the SCC
 * returns the selected brgrate for "real speed" calculation
 */
static unsigned int
scc_speed(register struct sccchan *scc,unsigned int clkmode,long speed)
{
	unsigned int brgrate;
	long spdclkm;
	int i_state;

	/* calculate baudrate generator value */

 	if ((spdclkm = speed * clkmode) == 0)
 	return 65000U;              	/* avoid divide-by-zero */

	brgrate = (unsigned) ((Sccinfo.clk + spdclkm) / (spdclkm * 2)) - 2;

 	i_state = dirps();          	/* 2-step register accesses... */

	cl(scc,R14,BRENABL);		/* disable baudrate generator */
	wr(scc,R12,brgrate);		/* brg rate LOW */
	wr(scc,R13,brgrate >> 8);   	/* brg rate HIGH */
	or(scc,R14,BRENABL);		/* enable baudrate generator */

	restore(i_state);

	return brgrate;
}

/*
 * Change scc_speed
 */
static void
scc_change_speed(unsigned int line)
{
	struct sccchan *scc;

	if((scc = Sccchan[line - SC_TTY_MASK]) == NULLCHAN)
		return;
	scc->speed = Sccinfo.clk / (64L * (scc_speed(scc,32,baud_table[scc->tty->termios->c_cflag & CBAUD]) + 2));
}



/* ----> ioctl-routine of the driver <---- */

/* perform ioctl on SCC (sdlc) channel
 * this is used for AX.25 mode, and will set the "kiss" parameters
 */
 
/* TIOCMGET - get modem status
 * TIOCMBIS - set PTT
 * TIOCMBIC - reset PTT
 * TIOCMBIC - set PTT
 * TIOCGSCC - get scc parameters
 * TIOCSSCC - set scc paremeters
 */
 
/* !!!!! Changes for patch level >= 9 !!!!! */

#ifdef oldlinux
static int
scc_ioctl(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned int arg)
#else
static int
scc_ioctl(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg)
#endif
{
	struct sccchan *scc;
	int i_state;
	unsigned int result;
	unsigned int value;
	struct scca * info;
	int error;

	if((scc = Sccchan[DEV_TO_SL(tty->line) - SC_TTY_MASK]) == NULLCHAN)
		return -EFAULT;
	if(!(scc->tty_opened))
		return -EFAULT;

	value = get_fs_long((unsigned long *) arg);
	info = (struct scca *) arg;

	switch(cmd){
	case TCSBRK:
		return 0;
	case TIOCMGET:

/* !!!!! Changes for patch level >= 9 !!!!! */

#ifdef oldlinux
		verify_area((void *) arg,sizeof(unsigned int *));
#else
		error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(unsigned int *));
		if (error)
			return error;
#endif		
		result =  ((scc->wreg[R5] & RTS) ? TIOCM_RTS : 0)
			| ((scc->wreg[R5] & DTR) ? TIOCM_DTR : 0)
			| ((RDREG(scc->ctrl) & DCD) ? TIOCM_CAR : 0)
			| ((RDREG(scc->ctrl) & CTS) ? TIOCM_CTS : 0);
		put_fs_long(result,(unsigned long *) arg);
		return 0;
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
		switch (cmd) {
		case TIOCMBIS:
			scc->wreg[R5] |= DTR;
			scc->wreg[R5] |= RTS;
			break;
		case TIOCMBIC:
			scc->wreg[R5] |= ~DTR;
			scc->wreg[R5] |= ~RTS;
			break;
		case TIOCMSET:
			if(value & TIOCM_DTR)
				scc->wreg[R5] |= DTR;
			else
				scc->wreg[R5] &= ~DTR;
			if(value & TIOCM_RTS)
				scc->wreg[R5] |= RTS;
			else
				scc->wreg[R5] &= ~RTS;
			break;
		}
		i_state = dirps();
		if(scc->a.tstate == IDLE && scc->timercount == 0)
		scc->timercount = 1;	/* force an update */
		restore(i_state);
		return 0;
	case TIOCGSCC:	

/* !!!!! Changes for patch level >= 9 !!!!! */

#ifdef oldlinux
		verify_area((void *) arg,sizeof(struct scca));
#else
		error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(struct scca));
		if (error)
			return error;
#endif
		
		if (!info)
			return -EFAULT;
		memcpy_tofs(info,&scc->a,sizeof(struct scca));
		return 0;
	case TIOCSSCC:
		if (!suser())
			return -EPERM;
		if (!info)
			return -EFAULT;
		memcpy_fromfs(&scc->a,info,sizeof(struct scca));
		return 0;
	default:
		return -EINVAL;
    }
}


/* ----- TERMIOS function ----- */

static void
scc_set_termios(struct tty_struct * tty, struct termios * old_termios)
{
	if (tty->termios->c_cflag == old_termios->c_cflag) 
		return;
	scc_change_speed(DEV_TO_SL(tty->line));
}



/*
 * this will set the "kiss" parameters through kiss itself
 */
 
static void
kiss_set_param(struct sccchan *scc,char cmd, unsigned int val)
{

#define VAL (val=val*TPS/100)? val:1

	switch(cmd){
	case PARAM_TXDELAY:
		scc->a.txdelay = VAL; break;
	case PARAM_PERSIST:
		scc->a.persist = val; break;
	case PARAM_SLOTTIME:
		scc->a.slottime = VAL; break;
	case PARAM_TXTAIL:
		scc->a.tailtime = VAL; break;
	case PARAM_FULLDUP:
		scc->a.fulldup = val; break;
	case PARAM_WAIT:
		scc->a.waittime = VAL; break;
	case PARAM_MAXKEY:
		scc->a.maxkeyup = val; break;
	case PARAM_MIN:
		scc->a.mintime = val; break;
	case PARAM_IDLE:
		scc->a.idletime = val; break;
	case PARAM_GROUP:
		scc->a.group = val;  break;
	case PARAM_TX:
		scc->a.tx_inhibit = val; 
	}
	return;
}

/* ----> tx routine: decode KISS data and enqueue it <---- */

/* send raw frame to SCC. used for AX.25 */
void
scc_write(struct tty_struct *tty)
{
	struct tty_queue *queue;
	struct sccchan *scc;
	unsigned char ch, kisscmd = 0;
	int i_state;

	if((scc = Sccchan[DEV_TO_SL(tty->line) - SC_TTY_MASK]) == NULLCHAN)
		return;

	queue = &scc->tty->write_q;

	if (scc->a.tx_inhibit){           /* transmitter inhibit */
		queue->tail = queue->head;
		return;
        }

	while(!(EMPTY(queue))){
		if(scc->sndq2 == NULLBUF){
			if(queue->buf[queue->tail] == FEND){
				queue->tail++;
				queue->tail &= TTY_BUF_SIZE-1;
			
/* !!!! ignore CRC on SMACK/RMNC-KISS-Frames !!!! */
			
				if ((kisscmd = queue->buf[queue->tail]) == FESC) {
					queue->tail++;
					queue->tail &= TTY_BUF_SIZE-1;
				
					kisscmd = queue->buf[queue->tail];
					if (kisscmd == TFEND)
						kisscmd = FEND;
					else
						kisscmd = FESC;
				}
			
				continue;
			}
			queue->tail++;
			queue->tail &= TTY_BUF_SIZE-1;
/* --- */			
						
			if((scc->sndq2 = scc->sndq1 = alloc_mbuf(scc->a.bufsiz)) == NULLBUF)
				return;
			continue;
		}
		if(scc->sndq2->cnt == scc->sndq2->size){        /* buffer full? */
			if((scc->sndq2 = alloc_mbuf(scc->a.bufsiz)) == NULLBUF)
				return;
			append(&scc->sndq1,scc->sndq2);         /* enqueue packet */
			continue;
		}

		ch = queue->buf[queue->tail++];
		queue->tail &= TTY_BUF_SIZE-1;
		switch(ch){
			case FEND:
			
/* !!!! strip the checksum and parse KISS-params !!!! */

				if (scc->sndq1->cnt && (kisscmd & 0xa0))
					if (scc->sndq1->cnt >= 2)
						scc->sndq1->cnt-=2;
					else
						scc->sndq1->cnt=0;

				kisscmd &= 0x1f;
				
				if (kisscmd && scc->sndq1->cnt)
				{
					kiss_set_param(scc, kisscmd, *scc->sndq1->data);
					scc->sndq1->cnt=0;
					
					free_p(scc->sndq1);
					scc->sndq2 = NULLBUF;
					break;
				}
/* --- */					
						
				if(scc->sndq1->cnt){
					enqueue(&scc->sndq,scc->sndq1);     /* enqueue packet */
					scc->sndq2 = NULLBUF;                   /* acquire a new buffer next time */
					scc->a.enqueued++;

					i_state = dirps();

					if(scc->a.tstate == IDLE){      /* when transmitter is idle */
						scc->a.tstate = DEFER;  /* start the key-up sequence */
							scc->a.maxdefer = TPS * scc->a.idletime /
							scc->a.slottime;
							if (scc->a.maxdefer == 0) 
                                   				scc->a.maxdefer = 400; /* dl1bke 940511: zero is very useless */

						scc->timercount = scc->a.waittime;
					}
					restore(i_state);
				}
				break;
			case FESC:
				if(!(EMPTY(queue))){
					ch = queue->buf[queue->tail++];
					queue->tail &= TTY_BUF_SIZE-1;
			
					switch(ch){
						case TFEND:
							scc->sndq2->data[scc->sndq2->cnt++] = FEND;
							break;
						case TFESC:
							scc->sndq2->data[scc->sndq2->cnt++] = FESC;
							break;
						default:
							break;
					}
				}
				else {
					queue->tail--;
					queue->tail &= TTY_BUF_SIZE-1;
					return;
				}
				break;
			default:
				scc->sndq2->data[scc->sndq2->cnt++] = ch;
				break;
		}
	}	
	return;
}


/* ----> Encode received data and write it to read_q  <---- */

/* receive raw frame from SCC. used for AX.25 */
static void
scc_rxraw(register struct sccchan *scc)
{
	struct mbuf *bp,*bp2;
	struct tty_queue * queue;
	int i_state;
	unsigned char ch;

	if(scc->rcvq1 == NULLBUF)
		return;

	queue = &scc->tty->read_q;

	while(!FULL(queue)){
		while ((bp = scc->rcvq1)->cnt == 0){      /* nothing left in this mbuf? */
			bp = scc->rcvq1->next;
			bp2 = scc->rcvq1->anext;
			i_state = dirps();
			free_mbuf(scc->rcvq1);    /*KM*/
			if((scc->rcvq1 = bp) == NULLBUF){/* see if more mbufs follow */
				scc->rcvq1 = bp2;
				restore(i_state);
				queue->buf[queue->head++] = FEND;
				queue->head &= TTY_BUF_SIZE-1;
				TTY_READ_FLUSH(scc->tty);
				queue->buf[queue->head++] = 0;
				queue->head &= TTY_BUF_SIZE-1;
				return;
			}
			restore(i_state);
		}
		switch(ch = (unsigned char)*(bp->data++)){
			case FEND:
				queue->buf[queue->head++] = FESC;
				queue->head &= TTY_BUF_SIZE-1;
				queue->buf[queue->head++] = TFEND;
				break;
			case FESC:
				queue->buf[queue->head++] = FESC;
				queue->head &= TTY_BUF_SIZE-1;
				queue->buf[queue->head++] = TFESC;
				break;
			default:
				queue->buf[queue->head++] = ch;
				break;
		}
		queue->head &= TTY_BUF_SIZE-1;

		bp->cnt--;
	}
	TTY_READ_FLUSH(scc->tty);
}

/* ----> Interrupt handlers for sdlc mode (AX.25) <---- */

/* Transmitter interrupt handler */
static void
scc_sdlctx(register struct sccchan *scc)
{
	register struct mbuf *bp;
	int x;

	scc->a.txints++;

	switch(scc->a.tstate){          /* look at transmitter state */
	case ACTIVE:                    /* busy sending data bytes */
		while ((bp = scc->tbp)->cnt == 0){      /* nothing left in this mbuf? */
			bp = bp->next;                  /* save link to next */
			free_mbuf(scc->tbp);    /*KM*/
			if((scc->tbp = bp) == NULLBUF){/* see if more mbufs follow */
				if(RDREG(scc->ctrl) & TxEOM){   /* check tx underrun status */
					scc->a.tovers++;          /* oops, an underrun! count them */
					WRREG(scc->ctrl,SEND_ABORT);/* send an abort to be sure */
#ifdef notdef
					scc->a.tstate = TAIL;   /* key down tx after TAILTIME */
					scc->timercount = scc->a.tailtime;
					return;
#else
					for(x=0;x!=1000;x++);
					WRREG(scc->ctrl,RES_EOM_L); /* reset the EOM latch */
					for(x=0;x!=1000;x++);
					WRREG(scc->ctrl,ERR_RES);
					for(x=0;x!=1000;x++);
					
					scc->a.tstate = FLUSH;   /* key down tx after TAILTIME */
					scc->timercount = scc->a.tailtime;
#endif
				}
				
				cl(scc,R10,ABUNDER);            /* frame complete, allow CRC transmit */
				scc->a.tstate = FLUSH;
				WRREG(scc->ctrl,RES_Tx_P);      /* reset pending int */
				return;
			}
		}
		/* now bp = scc->tbp (either from while or from if stmt above) */
		WRREG(scc->data,*(bp->data++)); /* send the character */
		bp->cnt--;                      /* decrease mbuf byte count */
		return;
	case FLUSH:     /* CRC just went out, more to send? */
#ifdef notdef	
		or(scc,R10,ABUNDER);            /* re-install underrun protection */
#endif
		/* verify that we are not exeeding max tx time (if defined) */
		if((scc->timercount != 0 || scc->a.maxkeyup == 0) &&
		 (scc->tbp = scc->sndq) != NULLBUF){ /* dequeue a frame */
			scc->sndq = scc->sndq->anext;
			WRREG(scc->ctrl,RES_Tx_CRC); /* reset the TX CRC generator */
			scc->a.tstate = ACTIVE;
			scc_sdlctx(scc);             /* write 1st byte */
			if(!(scc->enhanced))
				WRREG(scc->ctrl,RES_EOM_L);  /* reset the EOM latch */
			return;
		}
		scc->a.tstate = TAIL;           /* no more, key down tx after TAILTIME */
		scc->timercount = scc->a.tailtime;
		WRREG(scc->ctrl,RES_Tx_P);
		return;
	default:                                /* another state */
		WRREG(scc->ctrl,RES_Tx_P);      /* then don't send anything */
		return;
	}
}

/* External/Status interrupt handler */
static void
scc_sdlcex(register struct sccchan *scc)
{
	register unsigned char status,changes;

	scc->a.exints++;

	status = RDREG(scc->ctrl);
	changes = status ^ scc->status;

	if(changes & BRK_ABRT){         /* Received an ABORT */
		if(status & BRK_ABRT){          /* is this the beginning? */
			if(scc->rbp != NULLBUF){/* did we receive something? */
				/* check if a significant amount of data came in */
				/* this is because the drop of DCD tends to generate an ABORT */
				if(scc->rbp->next != NULLBUF || scc->rbp->cnt > 0)
				scc->a.rxerrs++;  /* then count it as an error */
				scc_tossb(scc);         /* throw away buffer */
			}
			VOID(RDREG(scc->data)); /* flush the FIFO */
			VOID(RDREG(scc->data));
			VOID(RDREG(scc->data));
		}
	}
	if(changes & CTS){                      /* CTS input changed state */
		if(status & CTS){               /* CTS is now ON */
			if(scc->a.tstate == KEYWT &&
				scc->a.txdelay == 0) /* zero TXDELAY = wait for CTS */
			scc->timercount = 1;    /* it will start within 10 ms */
		}
	}
	if(changes & DCD){                      /* DCD input changed state */
		if(status & DCD){               /* DCD is now ON */
			if (!scc->extclock)
				WRSCC(scc->ctrl,R14,SEARCH|scc->wreg[R14]); /* DPLL: enter search mode */
			or(scc,R3,ENT_HM|RxENABLE); /* enable the receiver, hunt mode */
		} else {                        /* DCD is now OFF */
			cl(scc,R3,ENT_HM|RxENABLE); /* disable the receiver */
			VOID(RDREG(scc->data)); /* flush the FIFO */
			VOID(RDREG(scc->data));
			VOID(RDREG(scc->data));
			if(scc->rbp != NULLBUF){/* did we receive something? */
				/* check if a significant amount of data came in */
				/* this is because some characters precede the drop of DCD */
				if(scc->rbp->next != NULLBUF || scc->rbp->cnt > 0)
				scc->a.rxerrs++;  /* then count it as an error */
				scc_tossb(scc);         /* throw away buffer */
			}
		}
	}
	scc->status = status;
	WRREG(scc->ctrl,RES_EXT_INT);
}

/* Receiver interrupt handler */
static void
scc_sdlcrx(register struct sccchan *scc)
{
	register struct mbuf *bp;

	scc->a.rxints++;

	if((scc->a.tstate == ACTIVE) && (scc->a.fulldup == 0))
	{
		VOID(RDREG(scc->data));	/* so we have to discard the char */
		or(scc,R3,ENT_HM);	/* enter hunt mode for next flag */
		return;
	}

	if((bp = scc->rbp1) == NULLBUF){ /* no buffer available now */
		if(scc->rbp == NULLBUF){
			if((bp = alloc_mbuf(scc->a.bufsiz)) != NULLBUF){
				scc->rbp = scc->rbp1 = bp;
			}
		} else if((bp = alloc_mbuf(scc->a.bufsiz)) != NULLBUF){
			scc->rbp1 = bp;
			for(bp = scc->rbp; bp->next != NULLBUF; bp = bp->next);
			bp->next = scc->rbp1;
			bp = scc->rbp1;
		}
		if(bp == NULLBUF){
			VOID(RDREG(scc->data)); /* so we have to discard the char */
			or(scc,R3,ENT_HM);      /* enter hunt mode for next flag */
			scc_tossb(scc);         /* put buffers back on pool */
			scc->a.nospace++;       /* count these events */
			return;
		}
	}

	/* now, we have a buffer (at bp). read character and store it */
	bp->data[bp->cnt++] = RDREG(scc->data);

	if(bp->cnt == bp->size)         /* buffer full? */
		scc->rbp1 = NULLBUF;    /* acquire a new one next time */
}


/* Receive Special Condition interrupt handler */
static void
scc_sdlcsp(register struct sccchan *scc)
{
	register unsigned char status;
	register struct mbuf *bp;

	scc->a.spints++;

	status = rd(scc,R1);            /* read receiver status */
	VOID(RDREG(scc->data));         /* flush offending character */

	if(status & Rx_OVR){            /* receiver overrun */
		scc->a.rovers++;                  /* count them */
		or(scc,R3,ENT_HM);              /* enter hunt mode for next flag */
		scc_tossb(scc);                 /* rewind the buffer and toss */
	}
	if(status & END_FR &&           /* end of frame */
	scc->rbp != NULLBUF){           /* at least received something */
		if((status & CRC_ERR) == 0 &&   /* no CRC error is indicated */
		(status & 0xe) == RES8 &&       /* 8 bits in last byte */
		scc->rbp->cnt){

			/* we seem to have a good frame. but the last byte received */
			/* from rx interrupt is in fact a CRC byte, so discard it */
			if(scc->rbp1 != NULLBUF){
				scc->rbp1->cnt--;       /* current mbuf was not full */
			} else {
				for(bp = scc->rbp; bp->next != NULLBUF; bp = bp->next);
					/* find last mbuf */

				bp->cnt--;              /* last byte is first CRC byte */
			}

			enqueue(&scc->rcvq1,scc->rbp);

			scc->rbp = scc->rbp1 = NULLBUF;
			scc->a.rxframes++;
		} else {                        /* a bad frame */
			scc_tossb(scc);         /* throw away frame */
			scc->a.rxerrs++;
		}
	}
	WRREG(scc->ctrl,ERR_RES);
}

/* Throw away receive mbuf(s) when an error occurred */
static void
scc_tossb(register struct sccchan *scc)
{
	register struct mbuf *bp;
	
	if((bp = scc->rbp) != NULLBUF){
		free_p(bp->next);
		free_p(bp->dup);        /* Should be NULLBUF */
		bp->next = NULLBUF;
		scc->rbp1 = bp;         /* Don't throw this one away */
		bp->cnt = 0;            /* Simply rewind it */
	}
}

/* ----> Switch the SCC to "transmit" mode <---- */

/* Only to be called from an interrupt handler, while in AX.25 mode */
static void
scc_txon(register struct sccchan *scc)
{
	if (!scc->fulldup && !scc->extclock){ /* no fulldup divider? */
		cl(scc,R3,RxENABLE);		/* then switch off receiver */
		cl(scc,R5,TxENAB);		/* transmitter off during switch */
		scc_speed(scc,1,scc->speed);	/* reprogram baudrate generator */

/* !!!! reprogram the Baudrategenerator !!!! */	
/* !!!! DL1BKE: Hmm, I think this is necessary for any card without
                a fullduplex divider. If you run into trouble, just
                remove the #ifdef/#endif constructions. See also scc_txoff() */
#ifdef 0
		if (Sccinfo.hwtype & HWBAYCOM){
#endif		
			/* DPLL -> Rx clk, BRG -> Tx CLK, TxCLK -> TRxC pin */
			wr(scc, R11, RCDPLL|TCBR|TRxCOI|TRxCBR);
#ifdef 0			
		}
#endif

/* --- */		
			
	}
	or(scc,R5,RTS|TxENAB);			/* set the RTS line and enable TX */

	if(Sccinfo.hwtype & HWPRIMUS)		/* PRIMUS has another PTT bit ... */ 
		WRREG(scc->ctrl + 4,Sccinfo.hwparam | 0x80);	/* set that bit ! */
}


/* ---->  Switch the SCC to "receive" mode (or: switch off transmitter) <---- */

/* Only to be called from an interrupt handler, while in AX.25 mod */

static void
scc_txoff(register struct sccchan *scc)
{
	cl(scc,R5,RTS);			/* turn off RTS line */
	if(Sccinfo.hwtype & HWPRIMUS)	/* PRIMUS has another PTT bit... */
		WRREG(scc->ctrl + 4,Sccinfo.hwparam);	/* clear that bit ! */

	if (!scc->fulldup && !scc->extclock){ /* no fulldup divider? */
		cl(scc,R5,TxENAB);		/* then disable the transmitter */
		scc_speed(scc,32,scc->speed);	/* back to receiver baudrate */
		
/* !!!! dito !!!! */
/* !!!! DL1BKE: see scc_txon() */

#ifdef 0
		if (Sccinfo.hwtype & HWBAYCOM){
#endif				
			/* DPLL -> Rx clk, DPLL -> Tx CLK, RxCLK -> TRxC pin */
			wr(scc, R11, RCDPLL|TCDPLL|TRxCOI|TRxCDP); 
#ifdef 0			
		}
#endif		
/* --- */
		
	}
}


/* ----> interrupt service routines for the 8530 <---- */

/* sccvec is the interrupt handler for SCC interrupts using INTACK
 */
static void
sccvec(int irq)
{
    register unsigned char vector;
    register struct sccchan *scc;
    register void (*handler)(register struct sccchan *scc);

    cli();
    /* Read SCC interrupt vector and check it */
    for(;;){ 
        outb(0,Sccvecloc);      /* Generate INTACK */
        /* scc_delay(5); */
        /* Read the vector */
        if((vector=inb(Sccvecloc)) >= Sccmaxvec)
          break;
        /* Extract channel number and status from vector. */
        /* Determine handler address. */
        /* Isolate channel nummer */
        if((vector & 0x01) || ((scc=Sccchan[(((vector>>1)&0x7c)^0x04) >> 2]) == NULLCHAN))
          break; 

        /* Isolate status info from vector, get handler, call handler */
        handler = (&scc->int_transmit)[(vector & 0x06) >> 1];
        handler(scc);                       /* Call the handler */
        outb(0x38,scc->ctrl);               /* Reset Highest IUS" opcode to WR0 */
        /* scc_delay(5); */
    }  
    sti();
}

/* sccnovec is the interrupt handler for SCC interrupts using polling
 */
static void
sccnovec(int irq)
{
    register unsigned char vector;
    register struct sccchan *scc;
    register void (*handler)(register struct sccchan *scc);
    register ioaddr *p;
    register ioaddr q;

    cli();
    /* Find the SCC generating the interrupt by polling all attached SCCs
     * reading RR3A (the interrupt pending register)
     */
     
    p = (unsigned short *) Sccpolltab;
    
    while(q=*(p++)){             
     outb(3,q);                         /* Read A CTRL address; Select RR3A */
     
     if(inb(q)){
        outb(2,q=*p);                   /* Read B CTRL address; Select RR2B */
        vector=inb(q);                  /* Read the vector */
        
        /* Extract channel number and status from vector. */
        /* Determine handler address. */
        /* Isolate channel nummer */
        if((vector & 1) || ((scc=Sccchan[(((vector>>1)&0x7c)^0x04) >> 2]) == NULLCHAN))
          break; 

        /* Isolate status info from vector, get handler, call handler */

        handler = (&scc->int_transmit)[(vector & 0x06) >> 1];
        handler(scc);
        p = (unsigned short *) Sccpolltab;
     }
    }    
    sti();
}


/* ----> SCC timer interrupt handler. Will be called every 1/TPS s <---- */

static void
scc_timer(void)
{
	register struct sccchan *scc;
	register struct sccchan **sccp;
	char i_state;

	struct tty_queue * queue;

	i_state = dirps();
	
	for(sccp = Sccchan + Sccinfo.maxchan; sccp >= Sccchan; sccp--){
		if((scc = *sccp) != NULLCHAN){
		  if(scc->timercount != 0 &&
		    --(scc->timercount) == 0){
			queue = &scc->tty->write_q;
			/* handle an SCC timer event for this SCC channel
			 * this can only happen when the channel is AX.25 type
			 */
			switch(scc->a.tstate){
			case IDLE:			/* it was idle, this is FULLDUP2 timeout */
				scc_txoff(scc);		/* switch-off the transmitter */
				break;
			case DEFER:			/* trying to get the channel */
				/* operation is as follows:
				 * CSMA: when channel clear AND persistence randomgenerator
				 *	 wins, AND group restrictions allow it:
				 *		keyup the transmitter
				 *	 if not, delay one SLOTTIME and try again
				 * FULL: always keyup the transmitter
				 */
				if(scc->a.fulldup == 0){
					Random = 21 * Random + 53;
					if(scc->status & DCD || scc->a.persist < Random){
						/* defer transmission again. check for limit */
defer_it:					if(--(scc->a.maxdefer) == 0){
							/* deferred too long. choice is to:
							 * - throw away pending frames, or
							 * - smash-on the transmitter and send them.
							 * the first would be the choice in a clean
							 * environment, but in the amateur radio world
							 * a distant faulty station could tie us up
							 * forever, so the second may be better...
							*/
#ifdef THROW_AWAY_AFTER_DEFER_TIMEOUT
#else
							goto keyup; /* just keyup the transmitter... */
#endif
						}
						scc->timercount = scc->a.slottime;
						break;
					}
					if(((unsigned char) scc->a.group) != NORTXGROUP){
						int i;
						struct sccchan *scc2;

						for(i = 0; i <= Sccinfo.maxchan; i++)
							if((scc2 = Sccchan[i]) != NULLCHAN &&
							 scc2 != scc &&
							 ((unsigned char) scc2->a.group) & ((unsigned char) scc->a.group) &&
							 ((scc->a.group & TXGROUP && scc2->wreg[R5] & RTS) ||
							 (scc->a.group & RXGROUP && scc2->status & DCD))){
								goto defer_it;
							}
					}
				}
			case KEYUP:			/* keyup transmitter (note fallthrough) */
keyup:				if((scc->wreg[R5] & RTS) == 0){ /* when not yet keyed */
					scc->a.tstate = KEYWT;
					scc->timercount = scc->a.txdelay; /* 0 if CTSwait */
					scc_txon(scc);
					break;
				}
				/* when already keyed, directly fall through */
			case KEYWT:                     /* waited for CTS or TXDELAY */
				/* when a frame is available (it should be...):
				 * - dequeue it from the send queue
				 * - reset the transmitter CRC generator
				 * - set a timeout on transmission length, if defined
				 * - send the first byte of the frame
				 * - reset the EOM latch
				 * when no frame available, proceed to TAIL handling
				 */
				if((scc->tbp = scc->sndq) != NULLBUF){
					scc->sndq = scc->sndq->anext;
					WRREG(scc->ctrl,RES_Tx_CRC);
					scc->a.tstate = ACTIVE;
					scc->timercount = TPS * scc->a.maxkeyup;
					scc_sdlctx(scc);
					if(!(scc->enhanced)) 
						WRREG(scc->ctrl,RES_EOM_L);
					break; 
				}
				/* when no frame queued, fall through to TAIL case */
			case TAIL:                      /* at end of frame */
				/* when fulldup is 0 or 1, switch off the transmitter.
				 * when frames are still queued (because of transmit time limit),
				 * restart the procedure to get the channel after MINTIME.
				 * when fulldup is 2, the transmitter remains keyed and we
				 * continue sending.    IDLETIME is an idle timeout in this case.
				 */     
				if(scc->a.fulldup < 2){
					scc->a.tstate = IDLE;
					scc_txoff(scc);

					if(scc->sndq != NULLBUF){
						scc->a.tstate = DEFER;
						scc->a.maxdefer = TPS * scc->a.idletime /
						 scc->a.slottime;
						if (scc->a.maxdefer == 0) scc->a.maxdefer = 400;
							/* dl1bke 940511: zero is not very useful */
 
						scc->timercount = TPS * scc->a.mintime;
					}
					break;
				}
				if(scc->sndq != NULLBUF){ /* still frames on the queue? */
					scc->a.tstate = KEYWT; /* continue sending */
					scc->timercount = TPS * scc->a.mintime; /* after mintime */
				} else {
					scc->a.tstate = IDLE;
					scc->timercount = TPS * scc->a.idletime;
				}
				break;
			case ACTIVE:	/* max keyup time expired */
			case FLUSH:	/* same while in flush mode */
				break;	/* no action required yet */
			default:			/* unexpected state */
				scc->a.tstate = IDLE;	/* that should not happen, but... */
				scc_txoff(scc);		/* at least stop the transmitter */
				break;
			}
		  }
		  i_state = dirps();
		  scc_rxraw(scc);
		  restore(i_state);
		}
	}
	timer_table[SCC_TIMER].fn = scc_timer;
	timer_table[SCC_TIMER].expires = jiffies + HZ/TPS;
	timer_active |= 1 << SCC_TIMER; 
	restore(i_state);
}
#endif
